(* $Id: godi_update.ml 1048 2012-07-09 15:48:51Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_exec
open Godi_query
open Godi_versions
open Godi_pkglist
open Printf


let available = ref PkgMap.empty
  (* maps wrapped packages to () *)


let wrapped_package (p:Godi_pkglist.pkg) build_tgz_name : wrapped_package =
  (* Returns a [wrapped_package] object for the build.tgz file with 
     absolute path [build_tgz_name]. The information from [p] is
     used.

     Note that dependency information for wrapped packages is not
     available (methods [dependencies], [build_dependencies], 
     [conflicts]), as deps are first calculated when the Makefiles are
     extracted.
   *)
  let (_, version, pkgrevision) = split_name_version ("x-" ^ p.pkg_version) in
  let var_id = Godi_types.new_var_id() in
  let ustatus = ref `Keep in
  let src_fingerprint_lz = lazy(
    "MD5:" ^ 
      Digest.to_hex(Digest.file(Godi_file.nativename build_tgz_name))) in
  ( object(self)
      method name = p.pkg_name
      method version = version
      method pkgrevision = pkgrevision
      method comment = p.pkg_comment
      method description = p.pkg_descr
      method godi_maintainer = p.pkg_maintainer
      method dependencies =
	failwith "dependencies: not available for wrapped package"
      method build_dependencies =
	failwith "build_dependencies: not available for wrapped package"
      method conflicts =
	failwith "conflicts: not available for wrapped package"
      method user_status = !ustatus
      method set_user_status u = ustatus := u
      method var_id = var_id
      method build_pkg = p
      method build_tarball = build_tgz_name
      method src_fingerprint =
	Some(Lazy.force src_fingerprint_lz)
    end
  )


let path_available_ini() =
  let bbase = Godi_file.filename(Godi_query.buildbase()) in
  let path = Godi_file.addname bbase "available.ini" in
  path


let read_available con =
  (* Read the list of locally available build.tgz ("wrapped packages")
     from disk, and update the [available] variable with this information.
   *)
  let bbase = Godi_file.filename(Godi_query.buildbase()) in
  let path = path_available_ini() in
  let npath = Godi_file.nativename path in 
  if Godi_file.file_exists path then (
    let l = Godi_pkglist.read_packages_ini npath in
    available := PkgMap.empty;
    List.iter
      (fun p ->
	 let build_tgz_name =
	   Godi_file.addname
	     (Godi_file.addname bbase "buildfiles")
	     (Godi_pkglist.build_tgz_name p) in
	 if Godi_file.file_exists build_tgz_name then (
	   let wp = `Wrapped_pkg (wrapped_package p build_tgz_name) in
	   available := PkgMap.add wp () !available
	 )
	 else 
	   printmsg con
	     ("Warning: file not found: " ^ 
		Godi_file.slashname build_tgz_name ^ "\n")
      )
      l;
  )
  else (
    printmsg con
      ("File not found: " ^ Godi_file.slashname path ^ "\n");
    printmsg con
      "(This is normal after an upgrade to godi-tools>=3.0, and will be fixed\n";
    printmsg con
      "the next time the packages are updated.)\n"
  )


let write_available con =
  (* Write the contents of [available] to disk *)
  let path = Godi_file.nativename (path_available_ini()) in
  let l = 
    List.map
      (fun wp -> (wrapped_pkg wp)#build_pkg)
      (PkgMap.keys !available) in
  Godi_pkglist.write_packages_ini path l


let words_re = Str.regexp "[ \t\r\n]+";;

let get_sites con =
  (* Get [sites], the list of GODI sites with build instructions: *)
  (* Caveat: profile.conf may also include additions to GODI_BUILD_SITES.
     This is taken into account here, but it only works for the 
     _current_ profile, not for an arbitrary or hypothetical one.
   *)
  let makefile = get_tools_makefile() in
  let b_sites = Buffer.create 100 in
  exec_to_user_error
    ~pref_user_error:true
    (call_cmd
       ~stdout:(`Buffer b_sites)
       ~stderr:(con.protocol :> Godi_exec.stderr_collector)
       ~interactivity:con.interactivity
       ~cd:(Godi_file.filename (buildbase())))
    ("godi_console", [| "make"; "-f"; makefile; "get_sites" |]);
  let sites = Str.split words_re (Buffer.contents b_sites) in
(* List.iter (fun s -> eprintf "%S\n%!" s) sites; *)
  sites


(*
let words_re = Str.regexp "[ \t\r\n]+";;

let parse_available con site file =
  let out = printmsg con in
  let f = 
    try
      Godi_file.open_in `Text file
    with
	Sys_error msg ->
	  user_error("System error: " ^ msg)
  in
  let l = ref [] in
  let n = ref 0 in
  ( try
      while true do
	incr n;
	let line = input_line f in
	let words = Str.split words_re line in
	match words with
	    [ name; dir; version ] ->
	      l := (name, (dir, version, site)) :: !l
	  | _ ->
	      out (sprintf "Warning: Parsing error in line %d of %s (site %s)\n"
		     !n (Godi_file.slashname file) site);
      done;
      assert false
    with
	End_of_file ->
	  close_in f;
      | error ->
	  close_in f;
	  raise error
  );
  (* Now [!l] are the packages from this site. Merge into 
   * available_current: 
   *)
  List.iter
    (fun (name, (dir, version, site)) ->
       let newer =
	 try
	   let (_, version', _) = Hashtbl.find available_current name in
	   let (_, v, rev)   = split_name_version ("x-" ^ version) in
	   let (_, v', rev') = split_name_version ("x-" ^ version') in
	   let d =
	     match cmp_version_string v v' with
		 0 -> rev - rev'
	       | n -> n in
	   d > 0
	 with
	     Not_found -> true in
       if newer then
	 Hashtbl.replace available_current name (dir, version, site)
    )
    !l;
  ()
;;
 *)

(*
let update_available con =
  let out = printmsg con in

  Hashtbl.clear available_current;
  let makefile = get_tools_makefile() in

  (* Get [sites], the list of GODI sites with build instructions: *)
  let b_sites = Buffer.create 100 in
  exec_to_user_error
    ~pref_user_error:true
    (call_cmd
       ~stdout:(`Buffer b_sites)
       ~stderr:(con.protocol :> Godi_exec.stderr_collector)
       ~interactivity:con.interactivity
       ~cd:(Godi_file.filename (buildbase())))
    ("godi_console", [| "make"; "-f"; makefile; "get_sites" |]);
  let sites = Str.split words_re (Buffer.contents b_sites) in

  (* For every site, download "available.new", and merge its
   * contents with the variable available_current
   *)
  List.iter
    (fun site ->
       (* Download available.new for [site]: *)
       exec_to_user_error
         ~pref_user_error:true
         (fun () ->
	    (* TODO: pass interactivity *)
	    let cleanup = ref (fun () -> ()) in
	    try
	      let hdl = Godi_file_handler.handler_of_directory site in
	      let (file, clean_opt) = hdl#get con "available.new" in
	      ( match clean_opt with
		  | Some f -> cleanup := f
		  | None -> ()
	      );
	      parse_available con site file;
	      !cleanup()
	    with
	      | File_handler_error msg ->
		  out ("Error: Cannot retrieve package list from " ^ site ^ ": " ^ msg ^ " (ignored)\n");
		  !cleanup()
	 )
         ();
    )
    sites
;;
 *)


let fetch_packages_ini profile con site =
  let out = printmsg con in
  let tmpdir = Godi_file.allocate_tmpdir() in
  try
    out ("===> Getting packages.ini from " ^ site ^ "\n");
(*
    let hdl = Godi_file_handler.handler_of_directory site in
    let (file, clean_opt) = hdl#get con "packages.ini" in
 *)
    let wtool = Wtool.wtool_factory() in
    ( match Godi_query.user() with
	| None -> ()
	| Some u -> wtool # configure_user u
    );
    let wlocation = wtool # open_location site in
    let basename =
      if profile.Godi_profile.prof_archive then
        "packages_all.ini"
      else
        "packages.ini" in
    let file_r = wlocation # get tmpdir ["/" ^ basename] in
    let file =
      match file_r with
	| [ `Ok file ] -> file
	| [ `Timeout ] -> failwith "Timeout"
	| [ `Error msg ] -> failwith msg
	| _ -> assert false in
    let pkgs =
      Godi_pkglist.read_packages_ini (Godi_file.nativename file) in
    Godi_file.release_tmpdir tmpdir;
    pkgs
  with
    |  Failure msg ->
	 out ("Error: Cannot retrieve package list from " ^ site ^ ": " ^ msg ^ " (ignored)\n");
	 Godi_file.release_tmpdir tmpdir;
	 []
    | error ->
	 Godi_file.release_tmpdir tmpdir;
	raise error


(* TODO:
   - also support a list of buildfiles, and download them in one go
     (pipelined)
 *)

let remote_name pkg =
  let pkgfullname = pkg.pkg_name ^ "-" ^ pkg.pkg_version in
  let remote_build_tgz_name = pkgfullname ^ ".build.tgz" in
  remote_build_tgz_name


let fetch_buildfiles con site (pkgs:Godi_pkglist.pkg list) = 
  let out = printmsg con in
  let tmpdir = Godi_file.allocate_tmpdir() in
  out ("===> Downloading build instructions from " ^ site ^ "\n");
  try
    let buildfiles = 
      Godi_file.addname (Godi_file.filename (buildbase())) "buildfiles" in
    
    Godi_file.create_dir ~recurse:true buildfiles;

    let wtool = Wtool.wtool_factory() in
    ( match Godi_query.user() with
	| None -> ()
	| Some u -> wtool # configure_user u
    );
    let wlocation = wtool # open_location site in
    let names = List.map remote_name pkgs in
    let absnames = List.map (fun s -> "/" ^ s) names in
    let files_r = wlocation # get tmpdir absnames in
    let have_error = ref false in
    List.iter2
      (fun pkg file_r ->
	 let name = remote_name pkg in
	 match file_r with
	   | `Ok file ->
	       ( try
		   let local_build_tgz_name = Godi_pkglist.build_tgz_name pkg in
		   let local_full_tgz_name = 
		     Godi_file.addname buildfiles local_build_tgz_name in
		   Godi_file.copy file local_full_tgz_name;
		 with
		   | error ->
		       out ("File " ^ name ^ ": Exception " ^ 
			      Netexn.to_string error);
		       have_error := true
	       )
	   | `Timeout -> 
	       out ("File " ^ name ^ ": Timeout"); have_error := true
	   | `Error msg -> 
	       out ("File " ^ name ^ ": " ^ msg); have_error := true
      )
      pkgs
      files_r;
    Godi_file.release_tmpdir tmpdir;
    if !have_error then
      raise(User_error ("Failed to download build instructions from: " ^
			  site));
  with
    | error ->
	Godi_file.release_tmpdir tmpdir;
	raise error


let subdir_re = Str.regexp "^\\([a-zA-Z0-9_]+\\)"
  

let extract_buildfile ?(tmp=false) con source pkg_name pkg_version build_tarball =
  (* tmp: extract into build/tmp *)
  (* pkg_version: includes pkgrevision *)
  let out = printmsg con in
  let cleanup = ref (fun () -> ()) in
  try
    let pkgfullname = pkg_name ^ "-" ^ pkg_version in
    if not tmp then
      out ("===> Extracting build instructions of " ^ pkgfullname ^ "\n");

    let subdir =
      if tmp then
	"tmp"
      else
	if Str.string_match subdir_re pkg_name 0 then
	  Str.matched_group 1 pkg_name
	else
	  "other" in

    let fulldir =
      Godi_file.addname (Godi_file.filename (buildbase())) subdir in

    Godi_file.create_dir ~recurse:true fulldir;

    let fulldir_pkgname = Godi_file.addname fulldir pkg_name in
    if Godi_file.file_exists fulldir_pkgname then
      Godi_file.remove_dir ~recurse:true fulldir_pkgname;

    cleanup := (fun () -> 
		  Godi_file.remove_dir ~recurse:true fulldir_pkgname);

    let ar = Pax.read (`File build_tarball) `Gzip `Tar in
    let cont = ref true in
    while !cont do
      match Pax.read_element ar with
	| None -> cont := false
	| Some el -> 
	    Pax.extract_element_as_file
	      ~prefix:fulldir
	      ~preserve_mtime:true
	      el
    done;
    Pax.postprocess_extraction ar;
    Pax.close_in ar;

    Godi_file.write_file_lines
      ~mode:`Binary
      (Godi_file.addname fulldir_pkgname ".source")
      [ source ];
    Godi_file.write_file_lines
      ~mode:`Binary
      (Godi_file.addname fulldir_pkgname ".version")
      [ pkg_version ];
    Godi_file.write_file_lines
      ~mode:`Binary
      (Godi_file.addname fulldir_pkgname ".src_fingerprint")
      [ "MD5:" ^ 
	  Digest.to_hex(Digest.file(Godi_file.nativename build_tarball)) ];

    cleanup := (fun () -> ());
  with
    | File_handler_error msg ->
	!cleanup();
	raise(User_error ("Failed to extract build instructions from: " ^
			    (Godi_file.slashname build_tarball) ^ 
			    " - error: " ^ msg))
    | error ->
	!cleanup();
	raise error
;;


let already_extracted source pkg_name pkg_version =
  let subdir =
    if Str.string_match subdir_re pkg_name 0 then
      Str.matched_group 1 pkg_name
    else
      "other" in
  let dir =
    Godi_file.addname (Godi_file.filename (buildbase())) subdir in
  let pkgdir = Godi_file.addname dir pkg_name in
  let source_file = Godi_file.addname pkgdir ".source" in
  let version_file = Godi_file.addname pkgdir ".version" in
  try
    let source_l = Godi_file.read_file_lines source_file in
    let version_l = Godi_file.read_file_lines version_file in
    source_l <> [] && version_l <> [] &&
      List.hd source_l = source &&
	List.hd version_l = pkg_version
  with
    | Sys_error _ ->
	false
  

let btgz_re = Str.regexp "^\\(.*--\\)?\\(.*\\)\\.build.tgz"
let src_re = Str.regexp "^\\(.*\\)--"


let name_vsuffix_of_build_name bname =
  if Str.string_match btgz_re bname 0 then (
    let name_version = Str.matched_group 2 bname in
    Godi_versions.split_name_vsuffix name_version
  )
  else
    failwith ("Godi_update: bad build filename: " ^ bname)
    

let source_of_build_name bname =
  if Str.string_match src_re bname 0 then
    Str.matched_group 1 bname
  else
    "public_godirepo_unknown"



let scan_buildfile con source build_tarball : Godi_pkglist.pkg =
  (* Extract meta information from a wrapped buildfile *)
  (* How to do: extract Makefile into the hierarchy: build/tmp/<name>/Makefile
     and then extract with Godi_query.query_makefile.
   *)
  let bname = Godi_file.basename build_tarball in
  let (name, vsuff) = name_vsuffix_of_build_name bname in
  extract_buildfile ~tmp:true con source name vsuff build_tarball;
  let dir =
    Godi_file.addname
      (Godi_file.addname (Godi_file.filename (buildbase())) "tmp")
      name in
  let vals =
    Godi_query.query_makefile
      ~uncached:true
      con
      dir
      [ "COMMENT"; "MAINTAINER"; "HOMEPAGE"; "DISTFILES" ] in
  let (comment, maintainer, homepage, distfiles) =
    match vals with
      | [ comment; maintainer; homepage; distfiles ] -> 
	  (comment, maintainer, homepage, distfiles)
      | _ -> assert false in
  let descr =
    Godi_query.read_text_file ~default:"" (Godi_file.addname dir "DESCR") in
  let pkg =
    { pkg_source = source;
      pkg_name = name;
      pkg_version = vsuff;
      pkg_comment = comment;
      pkg_descr = descr;
      pkg_maintainer = maintainer;
      pkg_homepage = homepage;
      pkg_svn_path = "";
      pkg_svn_rev = None;
      pkg_date = 0.0;
      pkg_rel_by = "";
      pkg_section = "unknown";
      pkg_distfiles = Str.split (Str.regexp "[ \t\r\b]+") distfiles;
    } in
  Godi_file.remove_dir ~recurse:true dir;
  pkg
;;

let scan_buildfile_dir con =
  (* Return the filenames of the available build.tgz files *)
  let buildfiles = 
    Godi_file.addname (Godi_file.filename (buildbase())) "buildfiles" in
  let names =
    if Godi_file.file_is_dir buildfiles then
      Godi_file.read_dir buildfiles
    else
      [] in
  let mnames =
    List.filter
      (fun n -> Str.string_match btgz_re n 0)
      names in
  mnames
;;


let group_by_name (pset : PkgSet.t) =
  (* returns a list of psets where each element contains all packages
     for a certain name
   *)
  let ht = Hashtbl.create 5 in
  PkgSet.iter
    (fun p ->
       let n = pkg_name p in
       let l = try Hashtbl.find ht n with Not_found -> PkgSet.empty in
       Hashtbl.replace ht n (PkgSet.add p l)
    )
    pset;
  Hashtbl.fold
    (fun _ l acc -> l :: acc)
    ht
    []


let best_in_pset (pset : PkgSet.t) =
  (* Returns the best package in pset *)
  let best_opt =
    PkgSet.fold
      (fun p acc ->
	 let better =
	   match acc with
	     | None -> true
	     | Some q -> newer_package p q in
	 if better then Some p else acc
      )
      pset
      None in
  match best_opt with
    | None -> raise Not_found
    | Some best -> best
    

let profile_best profile (pset : PkgSet.t) =
  (* Selects the best packages matching the profile conditions
     from the set [p]. A "best" package is a package with highest
     version number
   *)
  let pset1 =
    PkgSet.filter
      (fun p -> Godi_profile.profile_match profile (wrapped_pkg p))
      pset in
  let pset_groups = group_by_name pset1 in
  let best_groups = List.map best_in_pset pset_groups in
  let r =
    List.fold_left
      (fun acc p ->
	 PkgSet.add p acc
      )
      PkgSet.empty
      best_groups in
  r
;;


let source_of_pkgs pkgs =
  match pkgs with
    | [] -> raise Not_found
    | p :: _ -> p.pkg_source


let update_buildfiles con profile =
  (* (0) Ensure that [available] is initialized from the disk file.
     (1) Check whether there are buildfiles not registered in 
         [available]. If so, scan these buildfiles, and put their
         properties into [available]. Legacy buildfiles are renamed
         so they match the new naming scheme.
     (2) Update buildfiles from remote sites: 
         We first try to get packages.ini from the package source. If this
         works, the missing buildfiles are downloaded, and our local version
         of packages.ini is updated.

     Details for (1): The build.tgz files have names
     "<source>--<name>-<version>.build.tgz" (legacy: 
     "<name>-<version>.build.tgz").

     Details for (2): First, get the URLs of all build sites. Download
     packages.ini from each, but skip sources already encountered.
     (hhmm, define the source of a pkg list somehow). This gives us
     a big pkg list plus a mapping from source to URL. Filter the
     packages by profile (Godi_profile.filter). For each package name
     we only keep the package with the highest version matching the
     profile criterion. If this version is not in [available], we
     need to download the build.tgz (fetch_buildfile). Finally create 
     the wrapped package and add it to the variable [available].

   *)
  let buildfiles = 
    Godi_file.addname (Godi_file.filename (buildbase())) "buildfiles" in
  (* (0) *)
  read_available con;
  (* (1) *)
  let bnames = scan_buildfile_dir con in
  List.iter
    (fun bname ->
       let (n,vsuff) = name_vsuffix_of_build_name bname in
       let (v,rev) = Godi_versions.split_vsuffix vsuff in
       if not(PkgMap.mem_by 
		~name:n  ~version:v  ~pkgrevision:rev
		!available)
       then (
	 printmsg con ("===> Converting " ^ bname ^ "\n");
	 let src = source_of_build_name bname in
	 let build_tarball = Godi_file.addname buildfiles bname in
	 let pkg = scan_buildfile con src build_tarball in
	 let bname_new = Godi_pkglist.build_tgz_name pkg in
	 let build_tarball_new = Godi_file.addname buildfiles bname_new in
	 if bname_new <> bname then  (* rename file with legacy name *)
	   Godi_file.rename build_tarball build_tarball_new;
	 let wp = wrapped_package pkg build_tarball_new in
	 available := PkgMap.add (`Wrapped_pkg wp) () !available;
	 write_available con
       )
    )
    bnames;
  (* (2) *)
  let sites = get_sites con in   (* URLs of all build sites *)
  let sites_pkgs =
    List.rev
      (List.fold_left
	 (fun acc site ->
	    let pkgs1 = fetch_packages_ini profile con site in
	    let pkgs2 =
	      List.filter (fun pkg -> pkg.pkg_svn_rev <> None) pkgs1 in
	    if pkgs2 <> [] then
	      (site, pkgs2) :: acc
	    else
	      acc
	 )
	 []
	 sites
      ) in
  let sites_wps =
    List.map
      (fun (site, pkgs) ->
	 (* Create the (hypothetical) wrapped_package *)
	 let wps =
	   List.map
	     (fun pkg ->
		let build_tarball =
		  Godi_file.addname
		    buildfiles (Godi_pkglist.build_tgz_name pkg) in
		wrapped_package pkg build_tarball
	     )
	     pkgs in
	 (site, wps)
      )
      sites_pkgs in
  let dedup_ht = Hashtbl.create 5 in
    (* dedup: take only the first definition of a package from a certain
       source (in case there are several download possibilities)
     *)
  let all_wps =
    List.fold_left
      (fun acc (_, pkgs) -> 
	 List.fold_left
	   (fun acc pkg -> 
	      let source = (pkg#build_pkg).pkg_source in
	      let key = (pkg#name, pkg#version, pkg#pkgrevision, source) in
	      if not (Hashtbl.mem dedup_ht key) then (
		Hashtbl.add dedup_ht key ();
		PkgSet.add (`Wrapped_pkg pkg) acc
	      )
	      else acc
	   )
	   acc pkgs
      )
      PkgSet.empty
      sites_wps in
  let best_wps =
    profile_best profile all_wps in
  let best_missing_wps =
    PkgSet.filter
      (fun wp ->
	 let wp_source = ((wrapped_pkg wp)#build_pkg).pkg_source in
	 let l =
	   PkgMap.find_by 
	     ~name:(pkg_name wp)
	     ~version:(pkg_version wp)
	     ~pkgrevision:(pkg_pkgrevision wp)
	     !available in
	 not
	   (List.exists
	      (fun (ap,_) ->
		 ((wrapped_pkg ap)#build_pkg).pkg_source = wp_source
	      )
	      l
	   )
      )
      best_wps in
  let names = ref StrSet.empty in
  List.iter
    (fun (site, wps) ->
       let wps1 =
	 List.filter
	   (fun wp -> PkgSet.mem (`Wrapped_pkg wp) best_missing_wps)
	   wps in
       let wps_build =
	 List.map
	   (fun wp -> wp#build_pkg) 
	   wps1 in
       if wps_build <> [] then
	 fetch_buildfiles con site wps_build;
       List.iter
	 (fun wp ->
	    available := PkgMap.add (`Wrapped_pkg wp) () !available;
	    names := StrSet.add wp#name !names;
	 )
	 wps1;
       write_available con
    )
    sites_wps;
  !names
;;


let extract_buildfiles ?names con profile =
  (* Extract the right wrapped packages, and turn them into regular source
     packages.
   *)
  read_available con;
  let av = PkgMap.keys !available in
  let av_set = List.fold_left (fun acc p -> PkgSet.add p acc) PkgSet.empty av in
  let ex_set = profile_best profile av_set in
  let ex_sel =
    match names with
      | None -> ex_set
      | Some n_set ->
	  PkgSet.filter (fun p -> StrSet.mem (pkg_name p) n_set) ex_set in
  PkgSet.iter
    (fun p ->
       let v = 
	 Godi_versions.concat_vsuffix (pkg_version p) (pkg_pkgrevision p) in
       let build_tarball =
	 (wrapped_pkg p) # build_tarball in
       let source =
	 ((wrapped_pkg p) # build_pkg).pkg_source in
       if not (already_extracted source (pkg_name p) v) then
	 extract_buildfile con source (pkg_name p) v build_tarball  
    )
    ex_sel;
  Godi_query.forget_source_packages()
;;
